/* 
    qasn1 - Asn1BerNode - data structure class for qasn1 library
    Copyright(C) 2010 - Andrius da Costa Ribas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASN1BERNODE_H
#define ASN1BERNODE_H

/** @file Asn1BerNode.h
 * Header file for data structure class for qasn1 library
 */

#include "qasn1_export.h"

#include <QtCore/QByteArray>
#include <QtCore/QMetaType>

/**
 * Possible values of an ASN.1 Class ( bitmasks for the two first bits of tag portion }
 */
enum TagClasses
{
    CLASS_UNIVERSAL = 0, /**< UNIVERSAL */
    CLASS_APPLICATION = 0x40, /**< APPLICATION */
    CLASS_CONTEXT = 0x80, /**< CONTEXT-SPECIFIC */
    CLASS_PRIVATE = 0xC0 /**< PRIVATE */
};

/**
 * Possible values of an ASN.1 Coding ( bitmasks for the third bit of tag portion }
 */
enum CodingTypes
{
    CODING_PRIMITIVE = 0, /**< Primitive */
    CODING_COMPLEX = 0x20 /**< Complex (constructed) */
};

/**
 * Bitmasks for determining whether a tag is represented in more than one byte
 */
enum TagIndicators
{
    BIG_TAG_IND = 0x1F, /**< last five bits of first byte set to 1 means it is represented in more than one byte */
    BIG_TAG_NOT_LAST = 0x80, /**< bitmask for the first bit, indicating it is not the last byte used to represent the tag */
    BIG_TAG_LAST = 0 /**< bitmask for the first bit, indicating it is the last byte used to represent the tag */
};

/**
 * Miscellaneous bitmasks used in BER parsing
 */
enum BitMasks
{
    FIRST_BIT = 0x80, /**< First bit */
    TWO_FIRST_BITS = 0xC0, /**< Two first bits */
    THIRD_BIT = 0x20, /**< Third bit */
    LAST_FIVE_BITS = 0x1F, /**< Last five bits */
    LAST_SEVEN_BITS = 0x7F /**< Last seven bits */
};

/**
 * Bitmasks used in lenght parsing
 */
enum LengthIndicators
{
    BIG_LEN_IND = 0x80, /**< bitmask for the first bit of lenght octet to determine whether it is represented in more than one octet */
    UNKNOWN_LENGTH = 0x80 /**< value of lenght octet when it is in indefinite form */
};

class Asn1BerNodePrivate;

/**
 * Data structure to handle and ASN.1 (BER) Node
 */
class QASN1_EXPORT Asn1BerNode
{
private:
    
    Asn1BerNodePrivate* d;

public:
    
    /**
     * Default constructor, creates a single ASN.1 node, with undefined tag, no value and no related nodes
     */ 
    Asn1BerNode();
    /**
     * Constructor, creates a single ASN.1 node, with undefined tag, no value and specified parent node
     * @param parent pointer to the parent node
     */ 
    Asn1BerNode( Asn1BerNode* parent);
    /**
     * Copy constructor, copies the node specified in parameter "other", with no parent and copy of its entire subtree
     * @param other the node to be copied
     * @see clone()
     */     
    Asn1BerNode( const Asn1BerNode& other );
    /**
     * Destructor, removes all subnodes and re-link sibling nodes when necessary
     */
    virtual ~Asn1BerNode();
    /**
     * Clones the node
     * @return a copy of the node and its entire subtree
     * @see #Constructor(const Asn1BerNode&)
     */ 
    Asn1BerNode* clone();
    
    /**
     * Inserts node "other" before the calling node
     * @param other the node to be inserted
     */
    void insertBefore(Asn1BerNode* other);
    /**
     * Inserts node "other" after the calling node
     * @param other the node to be inserted
     */
    void insertAfter(Asn1BerNode* other);
    
    /**
     * Returns true if the calling node has a parent reference
     * @return <code>true</code> if the node has a parent node
     *         <code>false</code> if the node has no parent node
     */
    bool hasParent () const;
    /**
     * Returns true if the calling node has a child reference
     * @return <code>true</code> if the node has a child node
     *         <code>false</code> if the node has no child node
     */
    bool hasChild () const;
    /**
     * Returns true if the calling node has a sibling node after it
     * @return <code>true</code> if the node has a next node
     *         <code>false</code> if the node has no next node
     */
    bool hasNext () const;
    /**
     * Returns true if the calling node has a sibling node before it
     * @return <code>true</code> if the node has another node before
     *         <code>false</code> if the node has no other node before it
     */
    bool hasPrev () const;

    /**
     * Returns a pointer to the parent node
     * @return a pointer to the parent node, if present, or null pointer otherwise
     * @see hasParent()
     */
    Asn1BerNode* parent () const;
    /**
     * Returns a pointer to the first child node
     * @return a pointer to the child node, if present, or null pointer otherwise
     * @see hasChild()
     */
    Asn1BerNode* firstChild () const;
    /**
     * Returns a pointer to the following node
     * @return a pointer to the next node, if present, or null pointer otherwise
     * @see hasNext()
     */
    Asn1BerNode* next () const;
    /**
     * Returns a pointer to the previous node
     * @return a pointer to the previous node, if present, or null pointer otherwise
     * @see hasPrev()
     */
    Asn1BerNode* prev () const;
    
    /**
     * Sets N as the first child of the node, should be called only for first child linkage
     * @param N the child node
     * @see firstChild()
     */
    void setChild ( Asn1BerNode* N );
    /**
     * Links the calling node to its following node N
     * @param N the next node
     * @see next()
     */
    void setNext ( Asn1BerNode* N );
    /**
     * Links the calling node to its parent node N
     * @param N the parent node
     * @see parent()
     */
    void setParent ( Asn1BerNode* N );
    /**
     * Links the calling node to its previous node N
     * @param N the previous node
     * @see prev()
     */
    void setPrev ( Asn1BerNode* N );
    
    /**
     * Returns the ASN.1 tag number
     * @return the tag number
     * @see setTag(unsigned int)
     */
    unsigned int tag () const;
    /**
     * Returns the length of the array of data corresponding to the value assigned to the node
     * @return the length of the node's value
     * @see rawSize()
     */
    unsigned int len () const;
    /**
     * Returns a {@link QByteArray} corresponding to the value of the ASN.1 (BER) data
     * @return the value from ASN.1 node, or an empty {@link QByteArray} if it is a constructed node.
     * @see setVal(QByteArray)
     */
    QByteArray val () const;
    
    /**
     * Returns the raw size of the ASN.1 BER data, i.e.: the amount of bytes needed to represent the data in BER format
     * @return the raw size of the ASN.1 BER data
     * @see len()
     * @see Asn1BerSerializer
     */
    unsigned int rawSize () const;
    
    /**
     * Sets the tag number
     * @param tag the tag number to be assigned to the node
     * @see tag()
     */
    void setTag( unsigned int tag );
    /**
     * Sets the tag class
     * @param tag the tag class to be assigned to the node
     */
    void setClass( char tagClass );
    /**
     * Sets coding of the tag to primitive
     * @see isCodingPrimitive()
     * @see setCodingComplex()
     */
    void setCodingPrimitive();
    /**
     * Sets coding of the tag to complex (constructed)
     * @see isCodingComplex()
     * @see setCodingPrimitive()
     */
    void setCodingComplex();
    /**
     * Determines if the length octets are in definite or indefinite form ( only valid for constructed nodes ).
     * @param defined <code>true</code> if the lenght is in definite form, <code>false</code> otherwise
     * @see setCodingPrimitive()
     */
    void setDefinedLength(bool defined);
    /**
     * Sets the value of the node
     * @param val the raw value to be assigned to the node
     * @see setCodingPrimitive()
     */
    void setVal( QByteArray val );
    
    // methods to determine class of the Node
    /**
     * Whether a the tag class is of type UNIVERSAL
     * @return <code>true</code> if the class is of type UNIVERSAL, <code>false</code> otherwise
     * @see tagClass()
     */
    bool isClassUniversal () const;
    /**
     * Whether a the tag class is of type APPLICATION
     * @return <code>true</code> if the class is of type APPLICATION, <code>false</code> otherwise
     * @see tagClass()
     */
    bool isClassApplication () const;
    /**
     * Whether a the tag class is of type CONTEXT-SPECIFIC
     * @return <code>true</code> if the class is of type CONTEXT-SPECIFIC, <code>false</code> otherwise
     * @see tagClass()
     */
    bool isClassContext () const;
    /**
     * Whether a the tag class is of type PRIVATE
     * @return <code>true</code> if the class is of type PRIVATE, <code>false</code> otherwise
     * @see tagClass()
     */
    bool isClassPrivate () const;
    /**
     * Returns a char where the two first bits correspond to the tag class and the remaining bits are 0
     * @return one of the values of enum {@link TagClasses}
     * @see isClassUniversal()
     * @see isClassApplication()
     * @see isClassContext()
     * @see isClassPrivate()
     * @see TagClasses
     */
    char tagClass() const;

    // methods to determine Node coding
    /**
     * Whether the tag has coding primitive ( leaf nodes )
     * @return <code>true</code> if the tag coding is primitive, <code>false</code> otherwise
     * @see setCodingComplex()
     * @see isCodingPrimitive()
     */
    bool isCodingComplex () const;
    /**
     * Whether the tag has coding complex ( constructed )
     * @return <code>true</code> if the tag coding is constructed, <code>false</code> otherwise
     * @see setCodingPrimitive()
     * @see isCodingComplex()
     */
    bool isCodingPrimitive () const;
    
    /**
     * Whether constructed node has defined length ( always true for primitive nodes )
     * @return <code>true</code> if the lenght is in definite form, <code>false</code> otherwise
     * @see setDefinedLength(bool)
     */
    bool hasDefinedLength () const;
    
    /**
     * Returns the first node in the stream, i.e.: the node in top-level of the tree that has no previous node
     * @return the first node in the stream
     * @see last()
     */
    Asn1BerNode* first();
    /**
     * Returns the last node when navigating from the calling node, i.e. from it last sibling the last children, recursively
     * @return the last node
     * @see first()
     */
    Asn1BerNode* last();
    
    /**
     * Attaches an error message into the node, if the message is not empty the node becomes invalid
     * @param error the error message to be set
     * @see error()
     * @see isValid()
     */
    void setError(const QString& error);
    /**
     * Returns the error message attached to the node, if present
     * @return the error message or an empty {@link QString} if no error message present
     * @see setError()
     * @see isValid()
     */
    QString error() const;
    /**
     * Same of {@link error()}.isEmpty()
     * @see error()
     * @see setError()
     */
    bool isValid();

};

Q_DECLARE_METATYPE(Asn1BerNode)
Q_DECLARE_METATYPE(Asn1BerNode*)

#endif
